package de.halbekunst.juplo.cachecontrol;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.servlet.ModelAndView;

/**
 *
 * @author kai
 */
public class CacheControlInterceptor implements HandlerInterceptor {
  private final static Logger log = LoggerFactory.getLogger(CacheControlInterceptor.class);

  public static final String HEADER_DATE = "Date";
  public static final String HEADER_CACHE_CONTROL = "Cache-Control";
  public static final String HEADER_LAST_MODIFIED = "Last-Modified";
  public static final String HEADER_ETAG = "ETag";
  public static final String HEADER_EXPIRES = "Expires";
  public static final String HEADER_PRAGMA = "Pragma";
  public static final String HEADER_IF_MODIFIED_SINCE = "If-Modified-Since";
  public static final String HEADER_IF_NONE_MATCH = "If-None-Match";


  @Override
  public boolean preHandle(HttpServletRequest request, HttpServletResponse response, Object handler) throws Exception {
    try {
      Cacheable cacheable = (Cacheable)handler;

      long now = System.currentTimeMillis();

      /**
       * Alle Antworten (insbesondere auch 304) sollen nach dem {@plainlink
       * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.18 RFC 2616,
       * Abschnitt 14.18} einen Date-Header enthalten
       */
      response.setDateHeader(HEADER_DATE, now);

      /** Prüfen, ob der Handler willig ist, den Request zu verarbeiten */
      if (!cacheable.accepts(request)) {
        response.sendError(HttpServletResponse.SC_NOT_FOUND);
        return false;
      }

      /** Nichts weiter unternehmen, wenn der Handler dies nicht will */
      if (!cacheable.isGenerateCacheHeaders(request))
        return true;

      int cacheSeconds = cacheable.getCacheSeconds(request);
      if (cacheSeconds < 0) {
        response.setDateHeader(HEADER_DATE, now);
        response.setDateHeader(HEADER_EXPIRES, 0);
        response.addHeader(HEADER_PRAGMA, "no-cache");
        response.addHeader(HEADER_CACHE_CONTROL, "private");
        response.addHeader(HEADER_CACHE_CONTROL, "no-cache");
        response.addHeader(HEADER_CACHE_CONTROL, "no-store");
        response.addHeader(HEADER_CACHE_CONTROL, "max-age=0");
        response.addHeader(HEADER_CACHE_CONTROL, "s-max-age=0");
        return true;
      }

      long ifModifiedSince = -1;
      try {
        ifModifiedSince = request.getDateHeader(HEADER_IF_MODIFIED_SINCE);
      }
      catch (Exception e) {
        log.error("Exception while fetching If-Modified-Since: {}", e);
      }

      long lastModified = cacheable.getLastModified(request);

      /**
       * Sicherstellen, dass der Wert keine Millisekunden enthält, da die
       * Zeitangabe aus dem Modified-Since-Header keine Millisekunden enthalten
       * kann und der Test unten dann stets fehlschlagen würde!
       */
      lastModified = lastModified - (lastModified % 1000);

      String ifNoneMatch = request.getHeader(HEADER_IF_NONE_MATCH);
      String eTag = cacheable.getETag(request);

      /**
       * 304-Antworten sollen nach dem {@plainlink
       * http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.3.5 RFC
       * 2616, Abschnitt 10.3.5} einen ETag-Header enthalten, wenn auch die
       * 200-Antwort einen enthalten hätte.
       */
      if (eTag != null) {
        response.setHeader(HEADER_ETAG, eTag);
      }


      if (ifModifiedSince >= lastModified && lastModified > 0) {
        /**
         * request.getDateHeader liefert die Zeit als long, oder -1, wenn der
         * Header nicht existiert. D.h., wenn "If-Modified-Since" nicht gesetzt
         * ist, wird die komplette Seite ausgeliefert.
         * Der zusätzliche Test, ob lastModified größer 0 ist, ist nötig, um
         * Fehler auszuschließen, wenn die Implementierung von Cachable
         * negative Werte für Last-Modified zurückliefert.
         */
        if (log.isDebugEnabled())
          log.debug("Not modified since {}: {}", new Date(ifModifiedSince), request.getRequestURI());

        if (ifNoneMatch == null) {
          /** Neue Anfrage oder HTTP/1.0 Client! */
          log.debug("ETag nicht gesetzt: 304 {}", request.getRequestURI());
          response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
          return false;
        }
      }

      if (ifNoneMatch != null && ifNoneMatch.equals(eTag)) {
        log.debug("{}: ETag {} not changed -> 304 ", request.getRequestURI(), ifNoneMatch);
        response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
        return false;
      }

      /** HTTP/1.1-Caching-Header richtig setzen!! */
      response.setDateHeader(HEADER_LAST_MODIFIED, lastModified);

      /** Cache-Control für HTTP/1.1-Clients generieren */
      Map<String, String> cacheControl = new HashMap<String, String>(cacheable.getCacheControl(request));

      /**
       * Wenn eins JSESSIONID in der URL enthalten ist, darf die Anfrage nur vom
       * Browser gecached werden!
       */
      if (request.isRequestedSessionIdFromURL()) {
        cacheSeconds = 0;
        cacheControl.put("private", null);
      }

      if (cacheControl.containsKey("private")) {
        /**
         * HTTP/1.0 Caches davon abhalten, die Ressource zu cachen (vgl.: RFC
         * 2616, {@plainlink
         * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.9.3
         * Abschnitt 14.9.3} und {@plainlink
         * http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.32
         * Abschnitt 14.32})
         */
        response.setDateHeader(HEADER_EXPIRES, 0l);
        response.addHeader(HEADER_PRAGMA, "no-cache");
      }
      else {
        /**
         * Hier muss nicht geprüft werden, ob cacheSeconds > 0 gilt, da in diesem
         * Fall oben bereits No-Cache-Header generiert und <code>false</code>
         * zurückgeliefert werden!
         *
         * Den Wert als <code>max-age</code> zu den Schlüssel-Wert-Paaren für den
         * <code>Cache-Control</code>-Header hinzufügen und einen entsprechenden
         * <code>Expires</code>-Header für HTTP/1.0-Clients setzen.
         */
        cacheControl.put("max-age", Integer.toString(cacheSeconds));
        response.setDateHeader(HEADER_EXPIRES, (now + cacheSeconds * 1000));
      }

      StringBuilder builder = new StringBuilder();
      for (Entry<String, String> entry : cacheControl.entrySet()) {
        builder.setLength(0);
        builder.append(entry.getKey());
        if (entry.getValue() != null) {
          builder.append('=');
          builder.append(entry.getValue());
        }
        response.addHeader(HEADER_CACHE_CONTROL, builder.toString());
      }

      return true;
    }
    catch (ClassCastException e) {
      return true;
    }
  }

  @Override
  public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler, ModelAndView modelAndView) throws Exception {}

  @Override
  public void afterCompletion(HttpServletRequest request, HttpServletResponse response, Object handler, Exception ex) throws Exception {}
}
